# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: message.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import schema_pb2 as schema__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rmessage.proto\x12\x15relationalai.protocol\x1a\x0cschema.proto\"J\n\x0cMetadataInfo\x12:\n\trelations\x18\x01 \x03(\x0b\x32\'.relationalai.protocol.RelationMetadata\"]\n\x10RelationMetadata\x12\x36\n\x0brelation_id\x18\x01 \x01(\x0b\x32!.relationalai.protocol.RelationId\x12\x11\n\tfile_name\x18\x02 \x01(\tb\x06proto3')



_METADATAINFO = DESCRIPTOR.message_types_by_name['MetadataInfo']
_RELATIONMETADATA = DESCRIPTOR.message_types_by_name['RelationMetadata']
MetadataInfo = _reflection.GeneratedProtocolMessageType('MetadataInfo', (_message.Message,), {
  'DESCRIPTOR' : _METADATAINFO,
  '__module__' : 'message_pb2'
  # @@protoc_insertion_point(class_scope:relationalai.protocol.MetadataInfo)
  })
_sym_db.RegisterMessage(MetadataInfo)

RelationMetadata = _reflection.GeneratedProtocolMessageType('RelationMetadata', (_message.Message,), {
  'DESCRIPTOR' : _RELATIONMETADATA,
  '__module__' : 'message_pb2'
  # @@protoc_insertion_point(class_scope:relationalai.protocol.RelationMetadata)
  })
_sym_db.RegisterMessage(RelationMetadata)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _METADATAINFO._serialized_start=54
  _METADATAINFO._serialized_end=128
  _RELATIONMETADATA._serialized_start=130
  _RELATIONMETADATA._serialized_end=223
# @@protoc_insertion_point(module_scope)



# Copyright 2021 RelationalAI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers for formatting errors, problems and query results."""

import json
import sys
from typing import Union
from urllib.request import HTTPError

from railib.api import TransactionAsyncResponse

__all__ = ["http_error", "problems", "results"]


def http_error(e: HTTPError) -> None:
    rsp = json.loads(e.read())
    if rsp:
        print(f"status: {e.status}")
        print(json.dumps(rsp, indent=2))


def _show_row(row: list, end="\n"):
    row = [f'"{item}"' if isinstance(item, str) else str(item) for item in row]
    row = ", ".join(row)
    print(row, end=end)


# Print query response outputs as rel relations.
def _show_rel(rsp: dict) -> None:
    if rsp.get("aborted", False):
        print("aborted")
        return
    if rsp.get("output", False):
        outputs = rsp["output"]
        if len(outputs) == 0:
            print("false")
            return
        count = 0
        for output in outputs:
            cols = output["columns"]
            rkey = output["rel_key"]
            name = rkey["name"]
            keys = rkey["keys"]
            if name == "abort" and len(cols) == 0:
                continue  # ignore constraint results
            if count > 0:
                print()
            sig = "*".join(keys)
            print(f"// {name} ({sig})")
            rows = list(zip(*cols))
            if len(rows) == 0:
                print("true")
                continue
            for i, row in enumerate(rows):
                if i > 0:
                    print(";")
                _show_row(row, end="")
            print()
            count += 1
    if rsp.get("status", False):
        print(rsp["status"])


# Print the problems in the given response dict.
def problems(rsp: dict) -> None:
    if rsp is None:
        return
    problems = rsp.get("problems", None)
    if not problems:
        return
    for problem in problems:
        if problem.get("is_error", False):
            kind = "error"
        elif problem.get("is_exception", False):
            kind = "exception"
        else:
            kind = "warning"  # ?
        print(f"{kind}: {problem['message']}")
        report = problem.get("report", None)
        if report:
            print(report.rstrip())


# Print the results contained in the given response dict.
def results(rsp: Union[dict, list], format="physical") -> None:
    if rsp is None:
        return
    if format == "wire":
        json.dump(rsp, sys.stdout, indent=2)
    elif format == "physical":
        _show_rel(rsp)
        problems(rsp)
    else:
        raise Exception(f"unknown format: '{format}'")


# Print the results contained in the given TransactionAsyncResponse.
def results(rsp: TransactionAsyncResponse) -> None:
    if rsp.results is None:
        return

    for i, res in enumerate(rsp.results):
        print(res["relationId"])
        for v in zip(*res["table"].to_pydict().values()):
            print(v)

        if i < len(rsp.results) - 1:
            print()



# Copyright 2021 RelationalAI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
import time


__all__ = [
    "Credentials",
    "AccessToken",
    "ClientCredentials",
]

DEFAULT_CLIENT_CREDENTIALS_URL = "https://login.relationalai.com/oauth/token"


# Abstract base class shared by all credential types supported by railib.
class Credentials(ABC):
    pass


# Represents an OAuth access token.
class AccessToken:
    def __init__(self, access_token: str, scope: str, expires_in: int, created_on: float = time.time()):
        self.access_token = access_token
        self.scope = scope
        self.expires_in = expires_in
        self.created_on = created_on

    def is_expired(self):
        return (
            time.time() - self.created_on >= self.expires_in - 5
        )  # anticipate token expiration by 5 seconds


# Represents OAuth client credentials.
class ClientCredentials(Credentials):
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        client_credentials_url: str = None,
    ):
        self.access_token = None
        self.client_id = client_id
        self.client_secret = client_secret
        self.client_credentials_url = (
            client_credentials_url or DEFAULT_CLIENT_CREDENTIALS_URL
        )



# Copyright 2021 RelationalAI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reads the RAI SDK config file."""

# Note, the SDK config reader is here for convenience and its use is strictly
# optional. Many clients will chose to use other methods for managing settings
# and client credentials.

import configparser
import os
from pathlib import Path

from .credentials import ClientCredentials


def _read_config_profile(fname: str, profile: str) -> dict:
    config = configparser.ConfigParser()
    config.read(fname)
    if profile not in config:
        fname = os.path.basename(fname)
        raise Exception(f"profile '{profile}' not found in {fname}")
    return {k: config[profile][k] for k in config[profile]}


# Read client credentials from the config file. Returns None if they do not
# exist.
def _read_client_credentials(data):
    client_id = data.get("client_id", None)
    if client_id is not None:
        client_secret = data.get("client_secret", None)
        if client_secret is not None:
            client_credentials_url = data.get("client_credentials_url", None)
            return ClientCredentials(client_id, client_secret, client_credentials_url)
    return None


# Reads credentails from config file, preferring client credentials
# if they exist. Returns None if no credentials exist.
def _read_credentials(data, path):
    creds = _read_client_credentials(data)
    return creds


# Returns settings from config file.
def read(fname: str = "~/.rai/config", profile: str = "default"):
    path = Path(fname).expanduser()
    if not path.is_file():
        raise Exception(f"can't find file: {path}")
    data = _read_config_profile(path, profile)
    creds = _read_credentials(data, path)
    _keys = ["host", "port", "region", "scheme", "audience"]
    result = {k: v for k, v in data.items() if k in _keys}
    result["credentials"] = creds
    return result



# Copyright 2021 RelationalAI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Low level HTTP interface to the RelationalAI REST API."""

import json
import logging
from os import path
from urllib.parse import urlencode, urlsplit, quote
from urllib.request import Request, urlopen

from .__init__ import __version__
from .credentials import (
    AccessToken,
    Credentials,
    ClientCredentials,
)

__all__ = ["Context", "get", "put", "post", "request"]


ACCESS_KEY_TOKEN_KEY = "access_token"
CLIENT_ID_KEY = "client_id"
CLIENT_SECRET_KEY = "client_secret"
AUDIENCE_KEY = "audience"
GRANT_TYPE_KEY = "grant_type"
CLIENT_CREDENTIALS_KEY = "client_credentials"
EXPIRES_IN_KEY = "expires_in"
SCOPE = "scope"

# logger
logger = logging.getLogger(__package__)


# Context contains the state required to make rAI REST API calls.
class Context(object):
    def __init__(self, region: str = None, credentials: Credentials = None):
        self.region = region or "us-east"
        self.credentials = credentials
        self.service = "transaction"


# Answers if the keys of the passed dict contain a case insensitive match
# for the given term.
def _contains_insensitive(items: dict, term: str) -> bool:
    term = term.casefold()
    for item in items:
        item = item.casefold()
        if term == item:
            return True
    return False


# Fill in any missing headers.
def _default_headers(url: str, headers: dict = None) -> dict:
    headers = headers or {}
    if not _contains_insensitive(headers, "accept"):
        headers["Accept"] = "application/json"
    if not _contains_insensitive(headers, "content-type"):
        headers["Content-Type"] = "application/json"
    if not _contains_insensitive(headers, "host"):
        headers["Host"] = _get_host(url)
    if not _contains_insensitive(headers, "user-agent"):
        headers["User-Agent"] = _default_user_agent()
    return headers


def _default_user_agent() -> str:
    return f"rai-sdk-python/{__version__}"


def _encode(data) -> bytes:
    if not data:
        return data
    if not isinstance(data, str):
        data = json.dumps(data)
    return data.encode("utf8")


def _encode_path(path: str) -> str:
    # double encoding as per AWS v4
    return quote(quote(path))


# Returns an urlencoded query string.
# Note: the signing algo below **requires** query params to be sorted.
def _encode_qs(kwargs: dict) -> str:
    args = sorted([(k, v) for k, v in kwargs.items()])
    return urlencode(args)


# Retrieve the hostname from the given url.
def _get_host(url: str) -> str:
    return urlsplit(url).netloc.split(":")[0]


def _print_request(req: Request, level=0):
    if level <= 0:
        return
    if level > 0:
        print(f"{req.method} {req.full_url}")
        if level > 1:
            for k, v in req.headers.items():
                print(f"{k}: {v}")
            if req.data:
                print(json.dumps(json.loads(req.data.decode("utf8")), indent=2))


def _cache_file() -> str:
    return path.join(path.expanduser('~'), '.rai', 'tokens.json')


# Read oauth cache
def _read_cache() -> dict:
    try:
        with open(_cache_file(), 'r') as cache:
            return json.loads(cache.read())
    except Exception:
        return {}


# Read access token from cache
def _read_token_cache(creds: ClientCredentials) -> AccessToken:
    try:
        cache = _read_cache()
        return AccessToken(**cache[creds.client_id])
    except Exception:
        return None


# write access token to cache
def _write_token_cache(creds: ClientCredentials):
    try:
        cache = _read_cache()
        cache[creds.client_id] = creds.access_token

        with open(_cache_file(), 'w') as f:
            f.write(json.dumps(cache, default=vars))
    except Exception:
        pass


# Returns the current access token if valid, otherwise requests new token.
def _get_access_token(ctx: Context, url: str) -> AccessToken:
    creds = ctx.credentials
    assert isinstance(creds, ClientCredentials)
    if creds.access_token is None or creds.access_token.is_expired():
        creds.access_token = _read_token_cache(creds)
        if creds.access_token is None or creds.access_token.is_expired():
            creds.access_token = _request_access_token(ctx, url)
            _write_token_cache(creds)
    return creds.access_token.access_token


def _request_access_token(ctx: Context, url: str) -> AccessToken:
    creds = ctx.credentials
    assert isinstance(creds, ClientCredentials)
    # ensure the audience contains the protocol scheme
    audience = ctx.audience or f"https://{_get_host(url)}"
    headers = {
        "Accept": "application/json",
        "Content-Type": "application/json",
        "Host": _get_host(creds.client_credentials_url),
        "User-Agent": _default_user_agent(),
    }
    body = {
        CLIENT_ID_KEY: creds.client_id,
        CLIENT_SECRET_KEY: creds.client_secret,
        AUDIENCE_KEY: audience,
        GRANT_TYPE_KEY: CLIENT_CREDENTIALS_KEY,
    }
    data = _encode(body)
    req = Request(
        method="POST",
        url=creds.client_credentials_url,
        headers=headers,
        data=data,
    )
    _print_request(req)
    with urlopen(req) as rsp:
        result = json.loads(rsp.read())
        token = result.get(ACCESS_KEY_TOKEN_KEY, None)
        if token is not None:
            expires_in = result.get(EXPIRES_IN_KEY, None)
            scope = result.get(SCOPE, None)
            return AccessToken(token, scope, expires_in)
    raise Exception("failed to get the access token")


# Authenticate the request by signing or adding access token.
def _authenticate(ctx: Context, req: Request) -> Request:
    creds = ctx.credentials
    if creds is None:
        return req
    if isinstance(creds, ClientCredentials):
        access_token = _get_access_token(ctx, req.full_url)
        req.headers["authorization"] = f"Bearer {access_token}"
        return req
    raise Exception("unknown credential type")


# Issues an RAI REST API request, and returns response contents if successful.
def request(ctx: Context, method: str, url: str, headers={}, data=None, **kwargs):
    headers = _default_headers(url, headers)
    if kwargs:
        url = f"{url}?{_encode_qs(kwargs)}"
    data = _encode(data)
    req = Request(method=method, url=url, headers=headers, data=data)
    req = _authenticate(ctx, req)
    _print_request(req)
    rsp = urlopen(req)

    # logging
    content_type = headers["Content-Type"] if "Content-Type" in headers else ""
    agent = headers["User-Agent"] if "User-Agent" in headers else ""
    request_id = rsp.headers["X-Request-ID"] if "X-Request-ID" in rsp.headers else ""
    logger.debug(f"{rsp._method} HTTP/{rsp.version} {content_type} {rsp.url} {rsp.status} {agent} {request_id}")
    return rsp


def delete(ctx: Context, url: str, data, headers={}, **kwargs):
    return request(ctx, "DELETE", url, headers=headers, data=data, **kwargs)


def get(ctx: Context, url: str, headers={}, **kwargs):
    return request(ctx, "GET", url, headers=headers, **kwargs)


def put(ctx: Context, url: str, data, headers={}, **kwargs):
    return request(ctx, "PUT", url, headers=headers, data=data, **kwargs)


def post(ctx: Context, url: str, data, headers={}, **kwargs):
    return request(ctx, "POST", url, headers=headers, data=data, **kwargs)


def patch(ctx: Context, url: str, data, headers={}, **kwargs):
    return request(ctx, "PATCH", url, headers=headers, data=data, **kwargs)



import unittest

from railib import api


class TestPolling(unittest.TestCase):
    def test_timeout_exception(self):
        try:
            api.poll_with_specified_overhead(lambda: False, overhead_rate=0.1, timeout=1)
        except Exception as e:
            self.assertEqual('timed out after 1 seconds', str(e))

    def test_max_tries_exception(self):
        try:
            api.poll_with_specified_overhead(lambda: False, overhead_rate=0.1, max_tries=1)
        except Exception as e:
            self.assertEqual('max tries 1 exhausted', str(e))

    def test_validation(self):
        api.poll_with_specified_overhead(lambda: True, overhead_rate=0.1)
        api.poll_with_specified_overhead(lambda: True, overhead_rate=0.1, timeout=1)
        api.poll_with_specified_overhead(lambda: True, overhead_rate=0.1, max_tries=1)
        api.poll_with_specified_overhead(lambda: True, overhead_rate=0.1, timeout=1, max_tries=1)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 RelationalAI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Operation level interface to the RelationalAI REST API."""

import json
import pyarrow as pa
import time
import re
import io
import logging
from enum import Enum, unique
from typing import List, Union
from requests_toolbelt import multipart
from . import rest

from .pb.message_pb2 import MetadataInfo


PATH_ENGINE = "/compute"
PATH_DATABASE = "/database"
PATH_TRANSACTION = "/transaction"
PATH_TRANSACTIONS = "/transactions"
PATH_USER = "/users"
PATH_OAUTH_CLIENT = "/oauth-clients"

# logger
logger = logging.getLogger(__package__)


# Engine sizes
@unique
class EngineSize(str, Enum):
    XS = "XS"
    S = "S"
    M = "M"
    L = "L"
    XL = "XL"


# Database modes
@unique
class Mode(str, Enum):
    OPEN = "OPEN"
    CREATE = "CREATE"
    CREATE_OVERWRITE = "CREATE_OVERWRITE"
    OPEN_OR_CREATE = "OPEN_OR_CREATE"
    CLONE = "CLONE"
    CLONE_OVERWRITE = "CLONE_OVERWRITE"


# User roles
@unique
class Role(str, Enum):
    USER = "user"
    ADMIN = "admin"


# User/OAuth-client permissions
@unique
class Permission(str, Enum):
    # computes
    CREATE_COMPUTE = "create:compute"
    DELETE_COMPUTE = "delete:compute"
    LIST_COMPUTES = "list:compute"
    READ_COMPUTE = "read:compute"
    # engines (duplicate to computes)
    CREATE_ENGINE = "create:engine"
    DELETE_ENGINE = "delete:engine"
    LIST_ENGINES = "list:engine"
    READ_ENGINE = "read:engine"
    # databases
    LIST_DATABASES = "list:database"
    UPDATE_DATABASE = "update:database"
    DELETE_DATABASE = "delete:database"
    # transactions
    RUN_TRANSACTION = "run:transaction"
    READ_TRANSACTION = "read:transaction"
    # credits
    READ_CREDITS_USAGE = "read:credits_usage"
    # oauth clients
    CREATE_OAUTH_CLIENT = "create:oauth_client"
    READ_OAUTH_CLIENT = "read:oauth_client"
    LIST_OAUTH_CLIENTS = "list:oauth_client"
    UPDATE_OAUTH_CLIENT = "update:oauth_client"
    DELETE_OAUTH_CLIENT = "delete:oauth_client"
    ROTATE_OAUTH_CLIENT_SECRET = "rotate:oauth_client_secret"
    # users
    CREATE_USER = "create:user"
    LIST_USERS = "list:user"
    READ_USER = "read:user"
    UPDATE_USER = "update:user"
    # roles
    LIST_ROLES = "list:role"
    READ_ROLE = "read:role"
    # permissions
    LIST_PERMISSIONS = "list:permission"
    # access keys
    CREATE_ACCESS_KEY = "create:accesskey"
    LIST_ACCESS_KEYS = "list:accesskey"


__all__ = [
    "Context",
    "EngineSize",
    "Mode",
    "Role",
    "Permission",
    "create_database",
    "create_engine",
    "create_user",
    "create_oauth_client",
    "delete_database",
    "delete_engine",
    "delete_model",
    "disable_user",
    "enable_user",
    "delete_oauth_client",
    "get_database",
    "get_engine",
    "get_model",
    "get_oauth_client",
    "get_transaction",
    "get_transaction_metadata",
    "list_transactions",
    "get_transaction_results_and_problems",
    "cancel_transaction",
    "get_user",
    "list_databases",
    "list_edbs",
    "list_engines",
    "list_models",
    "list_users",
    "list_oauth_clients",
    "load_csv",
    "update_user",
]


# Context contains the state required to make rAI API calls.
class Context(rest.Context):
    def __init__(
        self,
        host: str = None,
        port: str = None,
        scheme: str = None,
        region: str = None,
        credentials=None,
        audience: str = None,
    ):
        super().__init__(region=region, credentials=credentials)
        self.host = host
        self.port = port or "443"
        self.scheme = scheme or "https"
        self.audience = audience


# Transaction async response class


class TransactionAsyncResponse:
    def __init__(
        self,
        transaction: dict = None,
        metadata: MetadataInfo = None,
        results: list = None,
        problems: list = None,
    ):
        self.transaction = transaction
        self.metadata = metadata
        self.results = results
        self.problems = problems

    def __str__(self):
        return str(
            {
                "transaction": self.transaction,
                "metadata": self.metadata,
                "results": self.results,
                "problems": self.problems,
            }
        )


# Transaction async file class


class TransactionAsyncFile:
    def __init__(
        self,
        name: str = None,
        filename: str = None,
        content_type: str = None,
        content: bytes = None,
    ):
        self.name = name
        self.filename = filename
        self.content_type = content_type
        self.content = content

    def __str__(self):
        return str(
            {
                "name": self.name,
                "filename": self.filename,
                "content-type": self.content_type,
                "content": "...",
            }
        )


# Construct a URL from the given context and path.


def _mkurl(ctx: Context, path: str) -> str:
    return f"{ctx.scheme}://{ctx.host}:{ctx.port}{path}"


# Retrieve an individual resource.
def _get_resource(ctx: Context, path: str, key=None, **kwargs) -> dict:
    url = _mkurl(ctx, path)
    rsp = rest.get(ctx, url, **kwargs)
    rsp = json.loads(rsp.read())
    if key:
        rsp = rsp[key]
    if rsp and isinstance(rsp, list):
        assert len(rsp) == 1
        return rsp[0]
    return rsp


# Retrieve a generic collection of resources.
def _get_collection(ctx, path: str, key=None, **kwargs):
    rsp = rest.get(ctx, _mkurl(ctx, path), **kwargs)
    rsp = json.loads(rsp.read())
    return rsp[key] if key else rsp


# Parse "multipart/form-data" response


def _parse_multipart_form(
    content_type: str, content: bytes
) -> List[TransactionAsyncFile]:
    result = []

    parts = multipart.decoder.MultipartDecoder(
        content_type=content_type, content=content
    ).parts

    for part in parts:
        txn_file = TransactionAsyncFile()
        txn_file.content_type = part.headers[b"Content-Type"].decode()
        txn_file.content = part.content

        disposition = part.headers[b"Content-Disposition"]
        name = re.match(b'.*; name="(.+?)"', disposition)
        if not (name is None):
            txn_file.name = name.group(1).decode()
        filename = re.match(b'.*filename="(.+?)"', disposition)
        if not (filename is None):
            txn_file.filename = name.group(1).decode()

        result.append(txn_file)

    return result


# Parse TransactionAsync response


def _parse_transaction_async_response(
    files: List[TransactionAsyncFile],
) -> TransactionAsyncResponse:
    txn_file = next(iter([file for file in files if file.name == "transaction"]), None)
    metadata_file = next(
        iter([file for file in files if file.name == "metadata.proto"]), None
    )
    problems_file = next(
        iter([file for file in files if file.name == "problems"]), None
    )

    if txn_file is None:
        raise Exception("transaction part is missing")
    if metadata_file is None:
        raise Exception("metadata.proto part is missing")
    if problems_file is None:
        raise Exception("problems part is missing")

    txn = json.loads(txn_file.content)
    metadata = _parse_metadata_proto(metadata_file.content)
    results = _parse_arrow_results(files)
    problems = json.loads(problems_file.content)

    return TransactionAsyncResponse(txn, metadata, results, problems)


# Parse Metadata from protobuf


def _parse_metadata_proto(data: bytes) -> MetadataInfo:
    metadata = MetadataInfo()
    metadata.ParseFromString(data)
    return metadata


# Extract arrow result from transaction async files


def _parse_arrow_results(files: List[TransactionAsyncFile]):
    results = []
    result_files = [
        file
        for file in files
        if file.content_type == "application/vnd.apache.arrow.stream"
    ]

    for file in result_files:
        with pa.ipc.open_stream(file.content) as reader:
            schema = reader.schema
            batches = [batch for batch in reader]
            table = pa.Table.from_batches(batches=batches, schema=schema)
            results.append({"relationId": file.name, "table": table})
    return results

# polling with specified overhead
# delay is the overhead % of the time the transaction has been running so far


def poll_with_specified_overhead(
    f,
    overhead_rate: float,
    start_time: int = time.time(),
    timeout: int = None,
    max_tries: int = None,
    max_delay: int = 120,
):
    tries = 0
    max_time = time.time() + timeout if timeout else None

    while True:
        if f():
            break

        if max_tries is not None and tries >= max_tries:
            raise Exception(f'max tries {max_tries} exhausted')

        if max_time is not None and time.time() >= max_time:
            raise Exception(f'timed out after {timeout} seconds')

        tries += 1
        duration = min((time.time() - start_time) * overhead_rate, max_delay)
        if tries == 1:
            time.sleep(0.5)
        else:
            time.sleep(duration)


def is_engine_term_state(state: str) -> bool:
    return state == "PROVISIONED" or ("FAILED" in state)


def create_engine(ctx: Context, engine: str, size: EngineSize = EngineSize.XS, **kwargs):
    data = {"region": ctx.region, "name": engine, "size": size.value}
    url = _mkurl(ctx, PATH_ENGINE)
    rsp = rest.put(ctx, url, data, **kwargs)
    return json.loads(rsp.read())


def create_engine_wait(ctx: Context, engine: str, size: EngineSize = EngineSize.XS, **kwargs):
    create_engine(ctx, engine, size, **kwargs)
    poll_with_specified_overhead(
        lambda: is_engine_term_state(get_engine(ctx, engine)["state"]),
        overhead_rate=0.2,
        timeout=30 * 60,
    )
    return get_engine(ctx, engine)


def create_user(ctx: Context, email: str, roles: List[Role] = None, **kwargs):
    rs = roles or []
    data = {"email": email, "roles": [r.value for r in rs]}
    url = _mkurl(ctx, PATH_USER)
    rsp = rest.post(ctx, url, data, **kwargs)
    return json.loads(rsp.read())


def create_oauth_client(ctx: Context, name: str, permissions: List[Permission] = None, **kwargs):
    ps = permissions or []
    data = {"name": name, "permissions": ps}
    url = _mkurl(ctx, PATH_OAUTH_CLIENT)
    rsp = rest.post(ctx, url, data, **kwargs)
    return json.loads(rsp.read())["client"]


# Derives the database open_mode based on the given arguments.
def _create_mode(source_database: str, overwrite: bool) -> Mode:
    if source_database is not None:
        result = Mode.CLONE_OVERWRITE if overwrite else Mode.CLONE
    else:
        result = Mode.CREATE_OVERWRITE if overwrite else Mode.CREATE
    return result


def delete_database(ctx: Context, database: str, **kwargs) -> dict:
    data = {"name": database}
    url = _mkurl(ctx, PATH_DATABASE)
    rsp = rest.delete(ctx, url, data, **kwargs)
    return json.loads(rsp.read())


def delete_engine(ctx: Context, engine: str, **kwargs) -> dict:
    data = {"name": engine}
    url = _mkurl(ctx, PATH_ENGINE)
    rsp = rest.delete(ctx, url, data, **kwargs)
    return json.loads(rsp.read())


def delete_user(ctx: Context, id: str, **kwargs) -> dict:
    url = _mkurl(ctx, f"{PATH_USER}/{id}")
    rsp = rest.delete(ctx, url, None, **kwargs)
    return json.loads(rsp.read())


def disable_user(ctx: Context, userid: str, **kwargs) -> dict:
    return update_user(ctx, userid, status="INACTIVE", **kwargs)


def delete_oauth_client(ctx: Context, id: str, **kwargs) -> dict:
    url = _mkurl(ctx, f"{PATH_OAUTH_CLIENT}/{id}")
    rsp = rest.delete(ctx, url, None, **kwargs)
    return json.loads(rsp.read())


def enable_user(ctx: Context, userid: str, **kwargs) -> dict:
    return update_user(ctx, userid, status="ACTIVE", **kwargs)


def get_engine(ctx: Context, engine: str, **kwargs) -> dict:
    return _get_resource(ctx, PATH_ENGINE, name=engine, deleted_on="", key="computes", **kwargs)


def get_database(ctx: Context, database: str, **kwargs) -> dict:
    return _get_resource(ctx, PATH_DATABASE, name=database, key="databases", **kwargs)


def get_oauth_client(ctx: Context, id: str, **kwargs) -> dict:
    return _get_resource(ctx, f"{PATH_OAUTH_CLIENT}/{id}", key="client", **kwargs)


def get_transaction(ctx: Context, id: str, **kwargs) -> dict:
    return _get_resource(ctx, f"{PATH_TRANSACTIONS}/{id}", key="transaction", **kwargs)


def get_transaction_metadata(ctx: Context, id: str, **kwargs) -> list:
    headers = {"Accept": "application/x-protobuf"}
    url = _mkurl(ctx, f"{PATH_TRANSACTIONS}/{id}/metadata")
    rsp = rest.get(ctx, url, headers=headers, **kwargs)
    content_type = rsp.headers.get("content-type", "")
    if "application/x-protobuf" in content_type:
        return _parse_metadata_proto(rsp.read())

    raise Exception(f"invalid content type for metadata proto: {content_type}")


def list_transactions(ctx: Context, **kwargs) -> list:
    return _get_collection(ctx, PATH_TRANSACTIONS, key="transactions", **kwargs)


def get_transaction_problems(ctx: Context, id: str, **kwargs) -> list:
    return _get_collection(ctx, f"{PATH_TRANSACTIONS}/{id}/problems", **kwargs)


def get_transaction_results(ctx: Context, id: str, **kwargs) -> list:
    url = _mkurl(ctx, f"{PATH_TRANSACTIONS}/{id}/results")
    rsp = rest.get(ctx, url, **kwargs)
    content_type = rsp.headers.get("content-type", "")
    if "multipart/form-data" in content_type:
        parts = _parse_multipart_form(content_type, rsp.read())
        return _parse_arrow_results(parts)

    raise Exception("invalid response type")


# When problems are part of the results relations, this function should be
# deprecated, get_transaction_results should be called instead


def get_transaction_results_and_problems(ctx: Context, id: str, **kwargs) -> list:
    rsp = TransactionAsyncResponse()
    rsp.problems = get_transaction_problems(ctx, id, **kwargs)
    rsp.results = get_transaction_results(ctx, id, **kwargs)
    return rsp


def cancel_transaction(ctx: Context, id: str, **kwargs) -> dict:
    rsp = rest.post(ctx, _mkurl(ctx, f"{PATH_TRANSACTIONS}/{id}/cancel"), {}, **kwargs)
    return json.loads(rsp.read())


def get_user(ctx: Context, userid: str, **kwargs) -> dict:
    return _get_resource(ctx, f"{PATH_USER}/{userid}", name=userid, **kwargs)


def list_engines(ctx: Context, state=None) -> list:
    kwargs = {}
    if state is not None:
        kwargs["state"] = state
    return _get_collection(ctx, PATH_ENGINE, key="computes", **kwargs)


def list_databases(ctx: Context, state=None) -> list:
    kwargs = {}
    if state is not None:
        kwargs["state"] = state
    return _get_collection(ctx, PATH_DATABASE, key="databases", **kwargs)


def list_users(ctx: Context, **kwargs) -> list:
    return _get_collection(ctx, PATH_USER, key="users", **kwargs)


def list_oauth_clients(ctx: Context, **kwargs) -> list:
    return _get_collection(ctx, PATH_OAUTH_CLIENT, key="clients", **kwargs)


def update_user(ctx: Context, userid: str, status: str = None, roles=None, **kwargs):
    data = {}
    if status is not None:
        data["status"] = status
    if roles is not None:
        data["roles"] = roles
    url = _mkurl(ctx, f"{PATH_USER}/{userid}")
    rsp = rest.patch(ctx, url, data, **kwargs)
    return json.loads(rsp.read())


class Transaction(object):
    def __init__(
        self,
        database: str,
        engine: str,
        abort=False,
        mode: Mode = Mode.OPEN,
        nowait_durable=False,
        readonly=False,
        source_database=None,
    ):
        self.abort = abort
        self.database = database
        self.engine = engine
        self.mode = mode
        self.nowait_durable = nowait_durable
        self.readonly = readonly
        self.source_database = source_database

    # Wrap each of the given actions in a LabeledAction structure.
    def _actions(self, *args):
        actions = []
        for i, action in enumerate(*args):
            actions.append(
                {
                    "name": f"action{i}",
                    "type": "LabeledAction",
                    "action": action,
                }
            )
        return actions

    @property
    def data(self):
        result = {
            "abort": self.abort,
            "dbname": self.database,
            "mode": self.mode.value,
            "nowait_durable": self.nowait_durable,
            "readonly": self.readonly,
            "type": "Transaction",
            "version": 0,
        }
        if self.engine is not None:
            result["computeName"] = self.engine
        if self.source_database is not None:
            result["source_dbname"] = self.source_database
        return result

    def run(self, ctx: Context, *args) -> dict:
        data = self.data
        data["actions"] = self._actions(args)
        # several of the request params are duplicated in the query
        kwargs = {
            "dbname": self.database,
            "compute_name": self.engine,
            "open_mode": self.mode.value,
            "region": ctx.region,
        }
        if self.source_database:
            kwargs["source_dbname"] = self.source_database
        url = _mkurl(ctx, PATH_TRANSACTION)
        rsp = rest.post(ctx, url, data, **kwargs)
        return json.loads(rsp.read())


class TransactionAsync(object):
    def __init__(
        self, database: str, engine: str, nowait_durable=False, readonly=False
    ):
        self.database = database
        self.engine = engine
        self.nowait_durable = nowait_durable
        self.readonly = readonly

    @property
    def data(self):
        result = {
            "dbname": self.database,
            "nowait_durable": self.nowait_durable,
            "readonly": self.readonly,
            # "sync_mode": "async"
        }
        if self.engine is not None:
            result["engine_name"] = self.engine
        return result

    def run(self, ctx: Context, command: str, language: str, inputs: dict = None, **kwargs) -> Union[dict, list]:
        data = self.data
        data["query"] = command
        data["language"] = language
        if inputs is not None:
            inputs = [_query_action_input(k, v) for k, v in inputs.items()]
            data["v1_inputs"] = inputs
        rsp = rest.post(ctx, _mkurl(ctx, PATH_TRANSACTIONS), data, **kwargs)
        content_type = rsp.headers.get("content-type", None)
        content = rsp.read()
        # todo: response model should be based on status code (200 v. 201)
        # async mode
        if content_type.lower() == "application/json":
            return json.loads(content)
        # sync mode
        if "multipart/form-data" in content_type.lower():
            return _parse_multipart_form(content_type, content)
        raise Exception("invalid response type")


def _delete_model_action(name: str) -> dict:
    return {"type": "ModifyWorkspaceAction", "delete_source": [name]}


def _install_model_action(name: str, model: str) -> dict:
    return {"type": "InstallAction", "sources": [_model(name, model)]}


def _list_action():
    return {"type": "ListSourceAction"}


def _list_edb_action():
    return {"type": "ListEdbAction"}


# Return rel key correponding to the given name and list of keys.
def _rel_key(name: str, keys: list) -> dict:
    return {"type": "RelKey", "name": name, "keys": keys, "values": []}


# Return the rel typename corresponding to the type of the given value.
def _rel_typename(v):
    if isinstance(v, str):
        return "RAI_VariableSizeStrings.VariableSizeString"
    raise TypeError("unsupported input type: {v.__class__.__name__")  # todo


# Return a qeury action input corresponding to the given name, value pair.
def _query_action_input(name: str, value) -> dict:
    return {
        "columns": [[value]],
        "rel_key": _rel_key(name, [_rel_typename(value)]),
        "type": "Relation",
    }


# `inputs`: map of parameter name to input value
def _query_action(model: str, inputs: dict = None, outputs: list = None) -> dict:
    inputs = inputs or {}
    inputs = [_query_action_input(k, v) for k, v in inputs.items()]
    return {
        "type": "QueryAction",
        "source": _model("query", model),
        "persist": [],
        "inputs": inputs,
        "outputs": outputs or [],
    }


def _model(name: str, model: str) -> dict:
    return {
        "type": "Source",
        "name": name,
        "path": "",  # todo: check if required?
        "value": model,
    }


# Returns full list of models.
def _list_models(ctx: Context, database: str, engine: str) -> dict:
    tx = Transaction(database, engine, mode=Mode.OPEN)
    rsp = tx.run(ctx, _list_action())
    actions = rsp["actions"]
    assert len(actions) == 1
    action = actions[0]
    models = action["result"]["sources"]
    return models


def create_database(ctx: Context, database: str, source: str = None, **kwargs) -> dict:
    data = {"name": database, "source_name": source}
    url = _mkurl(ctx, PATH_DATABASE)
    rsp = rest.put(ctx, url, data, **kwargs)
    return json.loads(rsp.read())


def delete_model(ctx: Context, database: str, engine: str, model: str) -> dict:
    tx = Transaction(database, engine, mode=Mode.OPEN, readonly=False)
    actions = [_delete_model_action(model)]
    return tx.run(ctx, *actions)


# Returns the named model
def get_model(ctx: Context, database: str, engine: str, name: str) -> str:
    models = _list_models(ctx, database, engine)
    for model in models:
        if model["name"] == name:
            return model["value"]
    raise Exception(f"model '{name}' not found")


def install_model(ctx: Context, database: str, engine: str, models: dict) -> dict:
    tx = Transaction(database, engine, mode=Mode.OPEN, readonly=False)
    actions = [_install_model_action(name, model) for name, model in models.items()]
    return tx.run(ctx, *actions)


def list_edbs(ctx: Context, database: str, engine: str) -> list:
    tx = Transaction(database, engine, mode=Mode.OPEN)
    rsp = tx.run(ctx, _list_edb_action())
    actions = rsp["actions"]
    assert len(actions) == 1
    action = actions[0]
    rels = action["result"]["rels"]
    return rels


# Returns a list of models installed in the given database.
def list_models(ctx: Context, database: str, engine: str) -> list:
    models = _list_models(ctx, database, engine)
    return [model["name"] for model in models]


# Generate a rel literal relation for the given dict.
def _gen_literal_dict(items: dict) -> str:
    result = []
    for k, v in items:
        result.append(f"{_gen_literal(k)},{_gen_literal(v)}")
    return "{" + ";".join(result) + "}"


# Generate a rel literal for the given list.
def _gen_literal_list(items: list) -> str:
    result = []
    for item in items:
        result.append(_gen_literal(item))
    return "{" + ",".join(result) + "}"


# Genearte a rel literal for the given string.
def _gen_literal_str(v: str) -> str:
    v = str(v)
    v = v.replace("'", "\\'")
    return f"'{v}'"


# Genrate a rel literal for the given value.
def _gen_literal(v) -> str:
    if isinstance(v, bool):
        return str(v).lower()
    if isinstance(v, str):
        return _gen_literal_str(v)
    if isinstance(v, dict):
        return _gen_literal_dict(v)
    if isinstance(v, list):
        return _gen_literal_list(v)
    return repr(v)


_syntax_options = ["header", "header_row", "delim", "quotechar", "escapechar"]


# Generate list of config syntax options for `load_csv`.
def _gen_syntax_config(syntax: dict = {}) -> str:
    result = ""
    for k in _syntax_options:
        v = syntax.get(k, None)
        if v is not None:
            result += f"def config:syntax:{k}={_gen_literal(v)}\n"
    return result


# `syntax`:
#   * header: a map from col number to name (base 1)
#   * header_row: row number of header, 0 means no header (default: 1)
#   * delim: default: ,
#   * quotechar: default: "
#   * escapechar: default: \
#
# Schema: a map from col name to rel type name, eg:
#   {'a': "int", 'b': "string"}
def load_csv(
    ctx: Context,
    database: str,
    engine: str,
    relation: str,
    data: str or io.TextIOBase,
    syntax: dict = {},
) -> dict:
    if isinstance(data, str):
        pass  # ok
    elif isinstance(data, io.TextIOBase):
        data = data.read()
    else:
        raise TypeError(f"bad type for arg 'data': {data.__class__.__name__}")
    inputs = {"data": data}
    command = _gen_syntax_config(syntax)
    command += "def config:data = data\n" "def insert:%s = load_csv[config]" % relation
    return exec_v1(ctx, database, engine, command, inputs=inputs, readonly=False)


def load_json(
    ctx: Context,
    database: str,
    engine: str,
    relation: str,
    data: str or io.TextIOBase,
) -> dict:
    if isinstance(data, str):
        pass  # ok
    elif isinstance(data, io.TextIOBase):
        data = data.read()
    else:
        raise TypeError(f"bad type for arg 'data': {data.__class__.__name__}")
    inputs = {"data": data}
    command = "def config:data = data\n" "def insert:%s = load_json[config]" % relation
    return exec_v1(ctx, database, engine, command, inputs=inputs, readonly=False)


def exec_v1(
    ctx: Context,
    database: str,
    engine: str,
    command: str,
    inputs: dict = None,
    readonly: bool = True,
) -> dict:
    tx = Transaction(database, engine, readonly=readonly)
    return tx.run(ctx, _query_action(command, inputs=inputs))


# Answers if the given transaction state is a terminal state.


def is_txn_term_state(state: str) -> bool:
    return state == "COMPLETED" or state == "ABORTED"


def exec(
    ctx: Context,
    database: str,
    engine: str,
    command: str,
    inputs: dict = None,
    readonly: bool = True,
    **kwargs
) -> TransactionAsyncResponse:
    logger.info('exec: database %s engine %s readonly %s' % (database, engine, readonly))
    start_time = time.time()
    txn = exec_async(ctx, database, engine, command, inputs=inputs, readonly=readonly)
    # in case of if short-path, return results directly, no need to poll for
    # state
    if not (txn.results is None):
        logger.info('transaction id: %s' % txn.transaction["id"])
        return txn

    rsp = TransactionAsyncResponse()
    txn = get_transaction(ctx, txn.transaction["id"], **kwargs)

    poll_with_specified_overhead(
        lambda: is_txn_term_state(get_transaction(ctx, txn["id"], **kwargs)["state"]),
        overhead_rate=0.2,
        start_time=start_time,
    )

    logger.info('transaction id: %s' % txn["id"])

    rsp.transaction = get_transaction(ctx, txn["id"], **kwargs)
    rsp.metadata = get_transaction_metadata(ctx, txn["id"], **kwargs)
    rsp.problems = get_transaction_problems(ctx, txn["id"], **kwargs)
    rsp.results = get_transaction_results(ctx, txn["id"], **kwargs)

    return rsp


def exec_async(
    ctx: Context,
    database: str,
    engine: str,
    command: str,
    readonly: bool = True,
    inputs: dict = None,
    language: str = "",
    **kwargs,
) -> TransactionAsyncResponse:
    tx = TransactionAsync(database, engine, readonly=readonly)
    rsp = tx.run(ctx, command, language=language, inputs=inputs, **kwargs)

    if isinstance(rsp, dict):
        return TransactionAsyncResponse(rsp, None, None, None)

    return _parse_transaction_async_response(rsp)


create_compute = create_engine  # deprecated, use create_engine
delete_compute = delete_engine  # deprecated, use delete_engine
get_compute = get_engine  # deprecated, use get_engine
list_computes = list_engines  # deprecated, use list_engines
list_edb = list_edbs  # deprecated, use list_edbs
delete_source = delete_model  # deprecated, use delete_model
get_source = get_model  # deprecated, use get_model
list_sources = list_models  # deprecated, use list_models



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: schema.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cschema.proto\x12\x15relationalai.protocol\"?\n\nRelationId\x12\x31\n\targuments\x18\x01 \x03(\x0b\x32\x1e.relationalai.protocol.RelType\"D\n\x08RelTuple\x12\x38\n\targuments\x18\x01 \x03(\x0b\x32%.relationalai.protocol.PrimitiveValue\".\n\tRelInt128\x12\x10\n\x08highbits\x18\x01 \x01(\x04\x12\x0f\n\x07lowbits\x18\x02 \x01(\x04\"/\n\nRelUInt128\x12\x10\n\x08highbits\x18\x01 \x01(\x04\x12\x0f\n\x07lowbits\x18\x02 \x01(\x04\"\xeb\x03\n\x0ePrimitiveValue\x12\x31\n\x03tag\x18\x01 \x01(\x0e\x32$.relationalai.protocol.PrimitiveType\x12\x36\n\nint128_val\x18\x02 \x01(\x0b\x32 .relationalai.protocol.RelInt128H\x00\x12\x13\n\tint64_val\x18\x03 \x01(\x03H\x00\x12\x13\n\tint32_val\x18\x04 \x01(\x05H\x00\x12\x13\n\tint16_val\x18\x05 \x01(\x05H\x00\x12\x12\n\x08int8_val\x18\x06 \x01(\x05H\x00\x12\x38\n\x0buint128_val\x18\x07 \x01(\x0b\x32!.relationalai.protocol.RelUInt128H\x00\x12\x14\n\nuint64_val\x18\x08 \x01(\x04H\x00\x12\x14\n\nuint32_val\x18\t \x01(\rH\x00\x12\x14\n\nuint16_val\x18\n \x01(\rH\x00\x12\x13\n\tuint8_val\x18\x0b \x01(\rH\x00\x12\x15\n\x0b\x66loat64_val\x18\x0c \x01(\x01H\x00\x12\x15\n\x0b\x66loat32_val\x18\r \x01(\x02H\x00\x12\x15\n\x0b\x66loat16_val\x18\x0e \x01(\x02H\x00\x12\x12\n\x08\x63har_val\x18\x0f \x01(\rH\x00\x12\x12\n\x08\x62ool_val\x18\x10 \x01(\x08H\x00\x12\x14\n\nstring_val\x18\x11 \x01(\x0cH\x00\x42\x07\n\x05value\"C\n\tValueType\x12\x36\n\x0e\x61rgument_types\x18\x01 \x03(\x0b\x32\x1e.relationalai.protocol.RelType\"p\n\x0c\x43onstantType\x12\x30\n\x08rel_type\x18\x01 \x01(\x0b\x32\x1e.relationalai.protocol.RelType\x12.\n\x05value\x18\x02 \x01(\x0b\x32\x1f.relationalai.protocol.RelTuple\"\xe3\x01\n\x07RelType\x12(\n\x03tag\x18\x01 \x01(\x0e\x32\x1b.relationalai.protocol.Kind\x12<\n\x0eprimitive_type\x18\x02 \x01(\x0e\x32$.relationalai.protocol.PrimitiveType\x12\x34\n\nvalue_type\x18\x03 \x01(\x0b\x32 .relationalai.protocol.ValueType\x12:\n\rconstant_type\x18\x04 \x01(\x0b\x32#.relationalai.protocol.ConstantType*\x92\x02\n\rPrimitiveType\x12\x14\n\x10UNSPECIFIED_TYPE\x10\x00\x12\x0b\n\x07INT_128\x10\x01\x12\n\n\x06INT_64\x10\x02\x12\n\n\x06INT_32\x10\x03\x12\n\n\x06INT_16\x10\x04\x12\t\n\x05INT_8\x10\x05\x12\x0c\n\x08UINT_128\x10\x06\x12\x0b\n\x07UINT_64\x10\x07\x12\x0b\n\x07UINT_32\x10\x08\x12\x0b\n\x07UINT_16\x10\t\x12\n\n\x06UINT_8\x10\n\x12\x0c\n\x08\x46LOAT_64\x10\x0b\x12\x0c\n\x08\x46LOAT_32\x10\x0c\x12\x0c\n\x08\x46LOAT_16\x10\r\x12\x08\n\x04\x43HAR\x10\x0e\x12\x08\n\x04\x42OOL\x10\x0f\x12\n\n\x06STRING\x10\x10\x12\n\n\x06SYMBOL\x10\x11\x12\x18\n\x14VARIABLE_SIZE_STRING\x10\x12*S\n\x04Kind\x12\x14\n\x10UNSPECIFIED_KIND\x10\x00\x12\x12\n\x0ePRIMITIVE_TYPE\x10\x01\x12\x0e\n\nVALUE_TYPE\x10\x02\x12\x11\n\rCONSTANT_TYPE\x10\x03\x62\x06proto3')

_PRIMITIVETYPE = DESCRIPTOR.enum_types_by_name['PrimitiveType']
PrimitiveType = enum_type_wrapper.EnumTypeWrapper(_PRIMITIVETYPE)
_KIND = DESCRIPTOR.enum_types_by_name['Kind']
Kind = enum_type_wrapper.EnumTypeWrapper(_KIND)
UNSPECIFIED_TYPE = 0
INT_128 = 1
INT_64 = 2
INT_32 = 3
INT_16 = 4
INT_8 = 5
UINT_128 = 6
UINT_64 = 7
UINT_32 = 8
UINT_16 = 9
UINT_8 = 10
FLOAT_64 = 11
FLOAT_32 = 12
FLOAT_16 = 13
CHAR = 14
BOOL = 15
STRING = 16
SYMBOL = 17
VARIABLE_SIZE_STRING = 18
UNSPECIFIED_KIND = 0
PRIMITIVE_TYPE = 1
VALUE_TYPE = 2
CONSTANT_TYPE = 3


_RELATIONID = DESCRIPTOR.message_types_by_name['RelationId']
_RELTUPLE = DESCRIPTOR.message_types_by_name['RelTuple']
_RELINT128 = DESCRIPTOR.message_types_by_name['RelInt128']
_RELUINT128 = DESCRIPTOR.message_types_by_name['RelUInt128']
_PRIMITIVEVALUE = DESCRIPTOR.message_types_by_name['PrimitiveValue']
_VALUETYPE = DESCRIPTOR.message_types_by_name['ValueType']
_CONSTANTTYPE = DESCRIPTOR.message_types_by_name['ConstantType']
_RELTYPE = DESCRIPTOR.message_types_by_name['RelType']
RelationId = _reflection.GeneratedProtocolMessageType('RelationId', (_message.Message,), {
  'DESCRIPTOR' : _RELATIONID,
  '__module__' : 'schema_pb2'
  # @@protoc_insertion_point(class_scope:relationalai.protocol.RelationId)
  })
_sym_db.RegisterMessage(RelationId)

RelTuple = _reflection.GeneratedProtocolMessageType('RelTuple', (_message.Message,), {
  'DESCRIPTOR' : _RELTUPLE,
  '__module__' : 'schema_pb2'
  # @@protoc_insertion_point(class_scope:relationalai.protocol.RelTuple)
  })
_sym_db.RegisterMessage(RelTuple)

RelInt128 = _reflection.GeneratedProtocolMessageType('RelInt128', (_message.Message,), {
  'DESCRIPTOR' : _RELINT128,
  '__module__' : 'schema_pb2'
  # @@protoc_insertion_point(class_scope:relationalai.protocol.RelInt128)
  })
_sym_db.RegisterMessage(RelInt128)

RelUInt128 = _reflection.GeneratedProtocolMessageType('RelUInt128', (_message.Message,), {
  'DESCRIPTOR' : _RELUINT128,
  '__module__' : 'schema_pb2'
  # @@protoc_insertion_point(class_scope:relationalai.protocol.RelUInt128)
  })
_sym_db.RegisterMessage(RelUInt128)

PrimitiveValue = _reflection.GeneratedProtocolMessageType('PrimitiveValue', (_message.Message,), {
  'DESCRIPTOR' : _PRIMITIVEVALUE,
  '__module__' : 'schema_pb2'
  # @@protoc_insertion_point(class_scope:relationalai.protocol.PrimitiveValue)
  })
_sym_db.RegisterMessage(PrimitiveValue)

ValueType = _reflection.GeneratedProtocolMessageType('ValueType', (_message.Message,), {
  'DESCRIPTOR' : _VALUETYPE,
  '__module__' : 'schema_pb2'
  # @@protoc_insertion_point(class_scope:relationalai.protocol.ValueType)
  })
_sym_db.RegisterMessage(ValueType)

ConstantType = _reflection.GeneratedProtocolMessageType('ConstantType', (_message.Message,), {
  'DESCRIPTOR' : _CONSTANTTYPE,
  '__module__' : 'schema_pb2'
  # @@protoc_insertion_point(class_scope:relationalai.protocol.ConstantType)
  })
_sym_db.RegisterMessage(ConstantType)

RelType = _reflection.GeneratedProtocolMessageType('RelType', (_message.Message,), {
  'DESCRIPTOR' : _RELTYPE,
  '__module__' : 'schema_pb2'
  # @@protoc_insertion_point(class_scope:relationalai.protocol.RelType)
  })
_sym_db.RegisterMessage(RelType)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PRIMITIVETYPE._serialized_start=1179
  _PRIMITIVETYPE._serialized_end=1453
  _KIND._serialized_start=1455
  _KIND._serialized_end=1538
  _RELATIONID._serialized_start=39
  _RELATIONID._serialized_end=102
  _RELTUPLE._serialized_start=104
  _RELTUPLE._serialized_end=172
  _RELINT128._serialized_start=174
  _RELINT128._serialized_end=220
  _RELUINT128._serialized_start=222
  _RELUINT128._serialized_end=269
  _PRIMITIVEVALUE._serialized_start=272
  _PRIMITIVEVALUE._serialized_end=763
  _VALUETYPE._serialized_start=765
  _VALUETYPE._serialized_end=832
  _CONSTANTTYPE._serialized_start=834
  _CONSTANTTYPE._serialized_end=946
  _RELTYPE._serialized_start=949
  _RELTYPE._serialized_end=1176
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python3
#
# Copyright 2021 RelationalAI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License

from setuptools import setup

import railib

setup(
    author="RelationalAI, Inc.",
    author_email="support@relational.ai",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="The RelationalAI Software Development Kit for Python",
    install_requires=[
        "ed25519==1.5",
        "pyarrow>=6.0.1",
        "requests-toolbelt==0.9.1",
        "protobuf==3.20.2"],
    license="http://www.apache.org/licenses/LICENSE-2.0",
    long_description="Enables access to the RelationalAI REST APIs from Python",
    long_description_content_type="text/markdown",
    name="rai-sdk",
    packages=["railib", "railib.pb"],
    url="http://github.com/RelationalAI/rai-sdk-python",
    version=railib.__version__)



import json
import unittest
import os
import uuid
import tempfile

from logging.config import fileConfig
from pathlib import Path
from railib import api, config


# Get creds from env vars if exists
client_id = os.getenv("CLIENT_ID")
client_secret = os.getenv("CLIENT_SECRET")
client_credentials_url = os.getenv("CLIENT_CREDENTIALS_URL")
host = os.getenv("HOST")
custom_headers = json.loads(os.getenv('CUSTOM_HEADERS', '{}'))

if client_id is None:
    print("using config from path")
    cfg = config.read()
else:
    file = tempfile.NamedTemporaryFile(mode="w")
    file.writelines(f"""
    [default]
    client_id={client_id}
    client_secret={client_secret}
    client_credentials_url={client_credentials_url}
    region=us-east
    port=443
    host={host}
    """)
    file.seek(0)
    cfg = config.read(fname=file.name)
    file.close()

ctx = api.Context(**cfg)

suffix = uuid.uuid4()
engine = f"python-sdk-{suffix}"
dbname = f"python-sdk-{suffix}"

# init "rai" logger
fileConfig("./test/logger.config")


class TestTransactionAsync(unittest.TestCase):
    def setUp(self):
        rsp = api.create_engine_wait(ctx, engine, headers=custom_headers)
        self.assertEqual("PROVISIONED", rsp["state"])
        rsp = api.create_database(ctx, dbname)
        self.assertEqual("CREATED", rsp["database"]["state"])

    def test_v2_exec(self):
        cmd = "x, x^2, x^3, x^4 from x in {1; 2; 3; 4; 5}"
        rsp = api.exec(ctx, dbname, engine, cmd)

        # transaction
        self.assertEqual("COMPLETED", rsp.transaction["state"])

        # metadata
        with open(os.path.join(Path(__file__).parent, "metadata.pb"), "rb") as f:
            data = f.read()
            self.assertEqual(
                rsp.metadata,
                api._parse_metadata_proto(data)
            )

        # problems
        self.assertEqual(0, len(rsp.problems))

        # results
        self.assertEqual(
            {
                'v1': [
                    1, 2, 3, 4, 5], 'v2': [
                    1, 4, 9, 16, 25], 'v3': [
                    1, 8, 27, 64, 125], 'v4': [
                        1, 16, 81, 256, 625]}, rsp.results[0]["table"].to_pydict())

    def tearDown(self):
        api.delete_engine(ctx, engine)
        api.delete_database(ctx, dbname)


if __name__ == '__main__':
    unittest.main()
